
/***************************************************************************
                          breakpoint.c  -  description
                             -------------------
    begin                : Sun Sep 16 2001
    copyright            : (C) 2001 by Keith Isdale
    email                : k_isdale@tpg.com.au
 ***************************************************************************/

/*
 * Implement a break point support. This is used primarily by functions
 *  in transform.c
 *
 * Several of the methods are not used yet
 */

#include "config.h"
#include "xsltutils.h"
#ifdef WITH_DEBUGGER
#include "breakpoint.h"
#include <string.h>

int xslDebugStatus = DEBUG_NONE;

#define MAX_BREAKPOINTS 40
#define MAX_TEMP_NAME_SIZE 40

xslBreakPoint breakList[MAX_BREAKPOINTS];

/* how many elements are valid*/
int breakIndex = 0;

/* keep track of what break point id we're up to*/
static breakpointCounter = 0;

xslCallPointPtr callStackBot, callStackTop;

xslCallPointInfoPtr callInfo;

/* What frame depth are we to stop at */
int stopDepth = -1;


/*
-----------------------------------------------------------
             Break point related functions
----------------------------------------------------------
*/


int activeBreakPoint = 0;

/**
 * xslFindTemplateNode: 
 * @style : valid stylesheet collection context to look into
 * @name : template name to look for
 *
 * Returns : template node found if successfull
 *           NULL otherwise with an error message will be printed if 
 *             WITH_XSLT_DEBUG_BREAKPOINTS is defined
 */
xmlNodePtr
xslFindTemplateNode(xsltStylesheetPtr style, const xmlChar * name)
{
    xmlNodePtr result = NULL;
    char *templName;
    xsltTemplatePtr templ;

    if (!style || !name) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Invalid stylesheet or template name : "
                         "xslFindTemplateNode\n");
#endif
        return result;
    }

    templ = style->templates;

    while (templ) {
        if (templ->match)
            templName = templ->match;
        else
            templName = templ->name;

        if (templName) {
            if (!strcmp(templName, name)) {
                result = templ->elem;
                break;
            }
        }
        templ = templ->next;
    }

#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
    if (!result)
        xsltGenericError(xsltGenericErrorContext,
                         "Template named '%s' not found :"
                         " xslFindTemplateNode\n", name);
#endif
    return result;
}


/**
 * xslFindBreakPointByLineNo:
 * @ctxt : valid ctxt to look into
 * @url : url non-null, non-empty file name that has been loaded by
 *                    debugger
 * @lineNumber : number >= 0 and is available in url specified
 *
 * Find the closest line number in file specified that can be a point 
 * Returns: line number number if successfull,
 *	    0 otherwise with an error message will be printed if 
 *             WITH_XSLT_DEBUG_BREAKPOINTS is defined
*/
xmlNodePtr
xslFindNodeByLineNo(xsltTransformContextPtr ctxt,
                    const xmlChar * url, long lineNumber)
{
    xmlNodePtr result = NULL;

    if (!ctxt || !url || (lineNumber == -1)) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Invalid ctxt, url or line number to "
                         "xslFindNodeByLineNo\n");
#endif
        return result;
    }

    return result;
}

/**
 * xslActiveBreakPoint();
 * Return the break point number that we stoped at
 */
int
xslActiveBreakPoint()
{
    return activeBreakPoint;
}

/**
 * xslSetActiveBreakPoint:
 * @breakPointNumber : 0 < breakPointNumber <= xslBreakPointCount()
 *
 * Return 1 on success o on failure
 */
int
xslSetActiveBreakPoint(int breakPointNumber)
{
    int result;

    if ((breakPointNumber < 0)
        || (breakPointNumber > xslBreakPointCount()))
        return result;

    activeBreakPoint = breakPointNumber;

    return ++result;
}


/**
 * xslAddBreakPoint:
 * @url : url non-null, non-empty file name that has been loaded by
 *                    debugger
 * @lineNumber : number >= 0 and is available in url specified and points to 
 *               an xml element
 * @temlateName : the template name of breakpoint or NULL
 * @type : DEBUG_BREAK_SOURCE if are we stopping at a xsl source line
 *         DEBUG_BREAK_DATA otherwise
 *
 * Add break point at file and line number specifiec
 * Returns: break point number if successfull,
 *	    0 otherwise with an error message will be printed if 
 *            WITH_XSLT_DEBUG_BREAKPOINTS is defined 
*/
int
xslAddBreakPoint(const xmlChar * url, long lineNumber,
                 const xmlChar * templateName, int type)
{

    int result = 0;

    if (!url || (lineNumber == -1)) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Invalid url or line number to xslAddBreakPoint\n");
#endif
        return result;
    }

    if (xslFindBreakPointByLineNo(url, lineNumber))
        return result;

    if (templateName && xslFindBreakPointByName(templateName))
        return result;


    if (breakIndex < MAX_BREAKPOINTS) {
        if (!xslIsBreakPoint(url, lineNumber)) {
            breakList[breakIndex].url = strdup(url);
            breakList[breakIndex].lineNo = lineNumber;
            if (templateName)
                breakList[breakIndex].templateName = strdup(templateName);
            else
                breakList[breakIndex].templateName = NULL;
            breakList[breakIndex].enabled = 1;
            breakList[breakIndex].id = ++breakpointCounter;
            breakList[breakIndex].type = type;
            breakIndex++;
            result = breakIndex;
        }
    } else {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Too many break points already: xslAddBreakPoint\n");
#endif
    }
    return result;
}


/**
 * xslDeleteBreakPoint:
 * @breakPointNumber : 0 < breakPointNumber <= xslBreakPointCount()
 *
 * Delete the break point specified
 * Returns: 1 if successfull,
 *	    0 otherwise with an error message will be printed if 
 *            WITH_XSLT_DEBUG_BREAKPOINTS is defined
*/
int
xslDeleteBreakPoint(int breakPointNumber)
{
    int index, result = 0;

    if ((breakPointNumber < 1)
        || (breakPointNumber > xslBreakPointCount())) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Break point number %d not found : xslDeleteBreakPoint\n",
                         breakPointNumber);
#endif
        return result;
    }

    breakPointNumber--;
    if (breakList[breakPointNumber].url) {
        free(breakList[breakPointNumber].url);
    }
    if (breakList[breakPointNumber].templateName) {
        free(breakList[breakPointNumber].templateName);
    }
    /* move all other elements down */
    for (index = breakPointNumber + 1; index < breakIndex; index++) {
        breakList[index - 1].url = breakList[index].url;
        breakList[index - 1].lineNo = breakList[index].lineNo;
        breakList[index - 1].templateName = breakList[index].templateName;
        breakList[index - 1].type = breakList[index].type;
        breakList[index - 1].id = breakList[index].id;
    }
    breakList[breakIndex - 1].url = NULL;
    breakList[breakIndex - 1].templateName = NULL;
    breakList[index - 1].id = -1;
    breakIndex--;
    result = 1;

    return result;
}


/**
 * xslFindBreakPointById:
 * @id : The break point id to look for
 *
 * Find the break point number for given break point id
 * Returns: break point number can be found for given the break point id
 *          0 otherwise with an error message printed if
 *            WITH_XSLT_DEBUG_BREAKPOINTS is defined
 */
int
xslFindBreakPointById(int id)
{
    int result = 0, index;

    for (index = 0; index < xslBreakPointCount(); index++) {
        if (breakList[index].id == id) {
            result = index + 1;
            break;
        }
    }

#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
    if (!result)
        xsltGenericError(xsltGenericErrorContext,
                         "Break point id %d not found :xslFindBreakPointById\n",
                         id);
#endif
    return result;
}


/**
 * xslFindBreakPointByLineNo:
 * @url :  a valid url that has been loaded by debugger
 * @lineNumber : lineNumber >= 0 and is available in url specified
 *
 * Find the break point number for a given url and line number
 * Returns: break point number number if successfull,
 *	    0 otherwise with an error message printed if 
 *             WITH_XSLT_DEBUG_BREAKPOINTS is defined
*/
int
xslFindBreakPointByLineNo(const xmlChar * url, long lineNumber)
{
    int result = 0, index;

    if (!url || (lineNumber == -1)) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Invalid url or line number to "
                         "xslFindBreakPointByLineNo\n");
#endif
        return result;
    }

    for (index = 0; index < xslBreakPointCount(); index++) {
        if ((breakList[index].lineNo == lineNumber) &&
            (!strcmp(breakList[index].url, url))) {
            result = index + 1;
            break;
        }
    }

    return result;
}


/**
 * xslFindBreakPointByName:
 * @templateName : template name to look for
 *
 * Find the breakpoint at template with "match" or "name" or equal to 
 *    templateName
 * Returns the break point number given the template name is found
 *          0 otherwise with an error message printed if 
 *             WITH_XSLT_DEBUG_BREAKPOINTS is defined
*/
int
xslFindBreakPointByName(const xmlChar * templateName)
{
    int result = 0, index;

    if (!templateName) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Template null :xslFindBreakPointByName\n");
#endif
        return result;
    }

    for (index = 0; index < xslBreakPointCount(); index++) {
        if (breakList[index].templateName &&
            !strcmp(breakList[index].templateName, templateName)) {
            result = index + 1;
            break;
        }
    }

    return result;
}


/**
 * xslEnableBreakPoint:
 * @breakPointNumber : 0 < breakPointNumber <= xslBreakPointCount()
 * @enable : enable break point if 1, disable if 0, toggle if -1
 *
 * Enable or disable a break point
 * Returns: 1 if successfull,
 *	    0 otherwise with an  error message printed if 
 *             WITH_XSLT_DEBUG_BREAKPOINTS is defined
*/
int
xslEnableBreakPoint(int breakPointNumber, int enable)
{
    int result = 0;

    if ((breakPointNumber < 1) || (breakPointNumber > breakIndex)) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Break point number %d not found : xslEnableBreakPoint\n",
                         breakPointNumber);
#endif
        return result;
    }

    breakPointNumber--;
    if (enable != XSL_TOGGLE_BREAKPOINT)
        breakList[breakPointNumber].enabled = enable;
    else {
        if (breakList[breakPointNumber].enabled)
            breakList[breakPointNumber].enabled = 0;
        else
            breakList[breakPointNumber].enabled = 1;
    }
    result++;
    return result;
}


/**
 * xslIsBreakPointEnabled:
 * @breakPointNumber : 0 < breakPointNumber <= xslBreakPointCount()
 *
 * Is the breakpoint at breakPointNumber specified enabled
 * Returns:  -1 if breakPointNumber is invalid
 *             with an error message will be printed if 
 *                WITH_XSLT_DEBUG_BREAKPOINTS is defined
 *           0 if break point is disabled 
 *           1 if break point is enabled      
*/
int
xslIsBreakPointEnabled(int breakPointNumber)
{
    int result = -1;

    if ((breakPointNumber < 1)
        || (breakPointNumber > xslBreakPointCount())) {
        /* This generates way to much text
         * #ifdef WITH_XSLT_DEBUG_BREAKPOINTS
         * xsltGenericError(xsltGenericErrorContext,
         * "Break point number %d not found : xslIsBreakPointEnabled\n", 
         * breakPointNumber);
         * #endif
         */
        return result;
    }

    breakPointNumber--;
    result = breakList[breakPointNumber].enabled;
    return result;
}


/**
 * xslBreakPointCount:
 *
 * Return the number of breakpoints present
 */
int
xslBreakPointCount()
{
    return breakIndex;
}


/**
 * xslLookupBreakPoint:
 * @breakPointNumber : 0 < breakPointNumber <= xslBreakPointCount()
 *
 * Lookup the value of break point at breakPointNumber specified
 * Returns: break point if breakPointNumber is valid,
 *	    NULL otherwise with an error message printed if 
 *             WITH_XSLT_DEBUG_BREAKPOINTS is defined
*/
xslBreakPointPtr
xslLookupBreakPoint(int breakPointNumber)
{
    xslBreakPointPtr breakPoint = NULL;

    if ((breakPointNumber < 1)
        || (breakPointNumber > xslBreakPointCount())) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Break point number %d not found : xslLookupBreakPoint\n",
                         breakPointNumber);
#endif
        return breakPoint;
    }
    breakPointNumber--;
    return &breakList[breakPointNumber];
}


/**
 * xslPrintBreakPoint:
 * @file : file != NULL
 * @breakPointNumber : 0 < breakPointNumber <= xslBreakPointCount()
 *
 * Print the details of break point to file specified
 *
 * Returns: 1 if successfull,
 *	    0 otherwise and error message will be printed if 
 *            WITH_XSLT_DEBUG_BREAKPOINTS is defined
 */
int
xslPrintBreakPoint(FILE * file, int breakPointNumber)
{
    int result = 0;

    if (!file || (breakPointNumber < 1)
        || (breakPointNumber > xslBreakPointCount())) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Break point number %d not found : xslPrintBreakPoint\n",
                         breakPointNumber);
#endif
        return result;
    }

    breakPointNumber--;
    fprintf(file, "Breakpoint %d ", breakList[breakPointNumber].id);
    if (breakList[breakPointNumber].enabled)
        fprintf(file, "enabled ");
    else
        fprintf(file, "disabled ");

    if (breakList[breakPointNumber].templateName) {
        fprintf(file, "for template :\"%s\" ",
                breakList[breakPointNumber].templateName);
    }

    if (breakList[breakPointNumber].url) {
        fprintf(file, "in file %s : line %ld",
                breakList[breakPointNumber].url,
                breakList[breakPointNumber].lineNo);
    } else {
        fprintf(file, "in file <n/a>, line %ld",
                breakList[breakPointNumber].lineNo);
    }
    return ++result;
}

/**
 * xslIsBreakPoint:
 * @url : url non-null, non-empty file name that has been loaded by
 *                    debugger
 * @lineNumber : number >= 0 and is available in url specified
 *
 * Determine if there is a break point at file and line number specifiec
 * Returns 1  if successfull,
 *	0 otherwise
*/
int
xslIsBreakPoint(const xmlChar * url, long lineNumber)
{
    int index, result = 0;

    if (!url || (lineNumber == -1))
        return result;

    result = xslFindBreakPointByLineNo(url, lineNumber);

    return result;
}


/**
 * xslIsBreakPointNode:
 * @node : node != NULL
 *
 * Determine if a node is a break point
 * Returns : 1 on sucess, 0 otherwise
 */
int
xslIsBreakPointNode(xmlNodePtr node)
{
    int index, result = 0;

    if (!node || !node->doc)
        return result;

    if (xmlGetLineNo(node) == -1)
        return result;

    if (node->doc->URL) {
        result = xslIsBreakPoint(node->doc->URL, xmlGetLineNo(node));
    }

    return result;
}


/*
-----------------------------------------------------------
       Main debugger functions
-----------------------------------------------------------
*/

/**
 * xslDebugInit :
 *
 * Initialize debugger
 */
void
xslDebugInit()
{
    int index;

    xslDebugStatus = DEBUG_CONT;
    /* stop at root node */
    /* add stuff to create break point structures */

    for (index = 0; index < MAX_BREAKPOINTS; index++) {
        breakList[index].url = NULL;
        breakList[index].lineNo = -1;
        breakList[index].id = -1;
        breakList[index].type = -1;
    }

    callInfo =
        (xslCallPointInfoPtr) xmlMemMalloc(sizeof(xslCallPointInfo));
    callInfo->next = NULL;
    callInfo->templateName = NULL;
    callInfo->url = NULL;
    callStackBot = (xslCallPointPtr) xmlMemMalloc(sizeof(xslCallPoint));
    callStackBot->next = NULL;
    callStackBot->info = NULL;
    callStackBot->lineNo = -1;
    callStackTop = callStackBot;
}


/**
 * xslDebugFree :
 *
 * Free up any memory taken by debugging
 */
void
xslDebugFree()
{
    int index;
    xslCallPointInfoPtr curInfo = callInfo, nextInfo;
    xslCallPointPtr curCall = callStackBot, nextCall;

    /* add stuff to free break point structures */

    /* remove all from file/line break point list */
    while (xslBreakPointCount())
        xslDeleteBreakPoint(1);

    /* remove all call info's */
    while (curInfo) {
        nextInfo = curInfo->next;
        if (curInfo->templateName)
            xmlMemFree(curInfo->templateName);
        if (curInfo->url)
            xmlMemFree(curInfo->url);
        curInfo = nextInfo;
    }

    /* remove all call stack items left. There should be none !! */
    while (curCall) {
        nextCall = curCall->next;
        xmlMemFree(curCall);
        curCall = nextCall;
    }
}


extern char *xslShellReadline(char *prompt);

/**
 * @templ : The source node being executed
 * @node : The data node being processed
 * @root : The template being applide to "node"
 * @ctxt : stylesheet being processed
 *
 * A break point has been found so pass control to user
 */
void
xslDebugBreak(xmlNodePtr templ, xmlNodePtr node, xsltTemplatePtr root,
              xsltTransformContextPtr ctxt)
{
    /*
     * if (root)
     * fprintf(stderr, "template :%s\n", root->match);
     * xslShell(templ, node, "index.xsl", xslShellReadline,stdout, ctxt);
     */
}



/*
------------------------------------------------------
                  Xsl call stack related
-----------------------------------------------------
*/

/**
 * xslAddCallInfo:
 * @templateName : template name to add
 * @url : url for templateName
 *
 * Returns a reference to the added info if sucessfull, otherwise NULL  
 */
xslCallPointInfoPtr
xslAddCallInfo(const xmlChar * templateName, const xmlChar * url)
{
    xslCallPointInfoPtr result = NULL, cur = callInfo;

    if (!templateName || !url) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Error : Null template name or url:"
                         " xslAddCallInfo\n");
        if (templateName)
            xsltGenericError(xsltGenericErrorContext, "template :'%s'\n",
                             templateName);
        if (url)
            xsltGenericError(xsltGenericErrorContext, "url :'%s'\n", url);

#endif
        return result;
    }

    while (cur->next) {
        if (cur->templateName && !strcmp(cur->templateName, templateName)
            && !strcmp(cur->url, url)) {
            result = cur;
            break;
        }
        cur = cur->next;
    }

    if (!result) {
        result =
            (xslCallPointInfoPtr) xmlMemMalloc(sizeof(xslCallPointInfo));
        if (result) {
            cur->next = result;
            result->templateName =
                (xmlChar *) xmlMemoryStrdup(templateName);
            result->url = (xmlChar *) xmlMemoryStrdup(url);
            result->next = NULL;
        } else {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
            xsltGenericError(xsltGenericErrorContext,
                             "Error : Unable to create xslCallPointInfo from :"
                             " xslAddCallInfo\n");
#endif
        }
    }
    return result;
}

/**
 * xslAddCall:
 * @templ : current template being applied
 * @source : the source node being processed
 *
 * Add template "call" to call stack
 * Returns : 1 on sucess 0 otherwise an error may be printed if 
 *            WITH_XSLT_DEBUG_BREAKPOINTS is defined
 */
int
xslAddCall(xsltTemplatePtr templ, xmlNodePtr source)
{
    int result = 0;
    const xmlChar *name = "Default template";
    xslCallPointInfoPtr info;

    if (!source->doc || !source->doc->URL) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Error : Invalid document url in call from : xslAddCall\n");
#endif
        return result;
    }


    /* this need not be an error just we've got a text in source */
    if (xmlGetLineNo(source) == -1) {
        return result;
    }

    if (templ) {
        if (templ->name)
            name = templ->name;
        else {
            if (templ->match)
                name = templ->match;
        }
    }

    if (!name) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Error : Invalid template name : xslAddCall\n");
#endif
        return result;
    }

    info = xslAddCallInfo(name, source->doc->URL);

    if (info) {
        xslCallPointPtr cur;

        cur = (xslCallPointPtr) xmlMemMalloc(sizeof(xslCallPoint));
        if (cur) {
            callStackTop->next = cur;
            callStackTop = cur;
            cur->info = info;
            cur->lineNo = xmlGetLineNo(source);
            cur->next = NULL;
            result++;
        } else {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
            xsltGenericError(xsltGenericErrorContext,
                             "Error : Unable to create call point : xslAddCall\n");
#endif
        }
    } else {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "Error : Unable to create call info : xslAddCall\n");
#endif
    }

    return result;
}

/** 
 * xslStepoutFromDepth :
 * @depth :call frame depth to step out from
 *
 * Set the depth to step out from
 */
void
xslStepoutFromDepth(int depth)
{
    stopDepth = depth;
}

/**
 * xslDropCall :
 *
 * Drop the topmost item off the call stack
 */
void
xslDropCall()
{

    if (!callStackBot) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "xslDropCall failed invalid call stack:"
                         " breakpoint.c");
#endif
        return;
    }

    /* are at a "frame" break point */
    if ((xslDebugStatus == DEBUG_STEPOUT) && stopDepth == xslCallDepth()) {
        xslDebugStatus = DEBUG_STOP;
        stopDepth = -1;
    }

    if (callStackBot->next) {
        xslCallPointPtr cur = callStackBot;

        while (cur->next && cur->next->next) {
            cur = cur->next;
        }
        if (cur->next)
            xmlMemFree(cur->next);
        cur->next = NULL;
        callStackTop = cur;
    } else {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "xslDropCall failed "
                         "no items on call stack : breakpoint.c");
#endif
    }
}


/**
 * xslGetCall :
 * @depth : 0 < depth <= xslCallDepth()  
 *
 * Retrieve the call point at specified call depth 

 * Return  non-null a if depth is valid
 *         NULL otherwise 
 */
xslCallPointPtr
xslGetCall(int depth)
{
    xslCallPointPtr result = NULL, cur = callStackBot;

    if (!callStackBot) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "xslGetCall failed invalid call stack:"
                         " breakpoint.c");
#endif
        return result;
    }
    if ((depth < 1) && (depth > xslCallDepth())) {
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "xslGetCall failed invalid call depth:"
                         " breakpoint.c");
#endif
        return result;
    }

    while (depth > 0 && cur->next) {
        cur = cur->next;
        depth--;
    }

    if (depth == 0)
        result = cur;
    else
#ifdef WITH_XSLT_DEBUG_BREAKPOINTS
        xsltGenericError(xsltGenericErrorContext,
                         "xslGetCall failed invalid call depth:"
                         " breakpoint.c");
#endif

    return result;
}


/** 
 * xslGetCallStackTop :
 *
 * Returns the top of the call stack
 */
xslCallPointPtr
xslGetCallStackTop()
{
    return callStackTop;
}


/** 
 * xslCallDepth :
 *
 * Return the depth of call stack
 */
int
xslCallDepth()
{
    xslCallPointPtr cur = callStackBot;
    int result = 0;

    while (cur->next) {
        result++;
        cur = cur->next;
    }
    return result;
}

#endif
